@@ -41,7 +41,7 @@ and workspace.
 """
 
 # Set version number
-__version__ = '4.6.1'
+__version__ = '4.7.0'
 
 import os
 import sys
@@ -62,6 +62,11 @@ if hasattr(sys, 'frozen') and sys.frozen:
         os.environ['FONTCONFIG_FILE'] = os.path.join(app_dir, 'source/pyzo/resources',
                                                     'fonts/linux_fonts.conf')
 
+# Automatically scale along on HDPI displays. See issue #531 and e.g.
+# https://wiki.archlinux.org/index.php/HiDPI#Qt_5
+if "QT_AUTO_SCREEN_SCALE_FACTOR" not in os.environ:
+    os.environ["QT_AUTO_SCREEN_SCALE_FACTOR"] = "1"
+
 # Import yoton as an absolute package
 from pyzo import yotonloader  # noqa
 from pyzo.util import paths
@@ -171,6 +176,29 @@ def resetConfig(preserveState=True):
     print("Deleted user config file. Restart Pyzo to revert to the default config.")
 
 
+def loadThemes():
+    """
+    Load default and user themes (if exist)
+    """
+
+    def loadThemesFromDir(dname, isBuiltin=False):
+        if not os.path.isdir(dname):
+            return
+        for fname in [fname for fname in os.listdir(dname) if fname.endswith(".theme")]:
+            try:
+                theme = ssdf.load(os.path.join(dname, fname))
+                assert theme.name.lower() == fname.lower().split(".")[0], "Theme name does not match filename"
+                theme.data = {key.replace("_", "."): val  for key, val in theme.data.items()}
+                theme["builtin"] = isBuiltin
+                themes[theme.name.lower()] = theme
+                print("Loaded theme %r" % theme.name)
+            except Exception as ex:
+                print("Warning ! Error while reading %s: %s" % (fname, ex))
+
+    loadThemesFromDir(os.path.join(pyzoDir, 'resources', 'themes'), True)
+    loadThemesFromDir(os.path.join(appDataDir, 'themes'))
+
+
 def loadConfig(defaultsOnly=False):
     """ loadConfig(defaultsOnly=False)
     Load default and site-wide configuration file(s) and that of the user (if it exists).
@@ -219,6 +247,7 @@ def loadConfig(defaultsOnly=False):
             print(t % fname)
             raise
 
+
 def saveConfig():
     """ saveConfig()
     Save all configureations to file.
@@ -237,6 +266,7 @@ def saveConfig():
         ssdf.save( os.path.join(appDataDir, "config.ssdf"), config )
 
 
+
 def start():
     """ Run Pyzo.
     """
@@ -293,5 +323,8 @@ _saveConfigFile = True
 config = ssdf.new()
 loadConfig()
 
+# Create style dict and fill it
+themes = {}
+loadThemes()
 # Init default style name (set in main.restorePyzoState())
 defaultQtStyleName = ''
